// TF has "azurerm_resource_provider_registration" ( https://registry.terraform.io/providers/hashicorp/azurerm/latest/docs/resources/resource_provider_registration )
// Pulumi Azure-Native, however, does not. So let's handle this via custom resources.
import * as pulumi from "@pulumi/pulumi";
import * as resources from "@azure/arm-resources";
import * as id from "@azure/identity";
export class ResourceProviderRegistration extends pulumi.dynamic.Resource {
    constructor(provider, name, args, opts) {
        opts = pulumi.mergeOptions(opts, {
            version: "1.0.0",
        });
        super(provider, `azure-custom:providers:Registrations:${name}`, args, opts);
    }
}
export class ResourceProviderRegistrationProvider {
    constructor(tenantId, subscriptionId, auth, keyAndCertPath) {
        this.tenantId = tenantId;
        this.subscriptionId = subscriptionId;
        this.auth = auth;
        this.keyAndCertPath = keyAndCertPath;
    }
    async create(inputs) {
        const client = getClient(this);
        await Promise.all(inputs.resourceProviderNamespaces.map((resourceProviderNamespace) => client.providers.register(resourceProviderNamespace)));
        return {
            id: getPulumiResourceID(this),
            outs: inputs,
        };
    }
    check(olds, news) {
        const failures = [];
        if (news.resourceProviderNamespaces.length <= 0) {
            failures.push({
                property: "resourceProviderNamespaces",
                reason: "There must be at least one resource provider namespace specified",
            });
        }
        return Promise.resolve(failures.length > 0
            ? {
                failures,
            }
            : {
                inputs: news,
            });
    }
    diff(id, { resourceProviderNamespaces: nsOut }, { resourceProviderNamespaces: nsIn }) {
        const providersAreSame = nsIn.length === nsOut.length &&
            nsIn.every((nsi) => nsOut.some((nso) => nsi.toLowerCase() == nso.toLowerCase()));
        return Promise.resolve({
            // We never replace this resource
            deleteBeforeReplace: false,
            replaces: [],
            changes: !providersAreSame,
        });
    }
    async read(id, currentProps) {
        const providers = await getClient(this).providers.list();
        return {
            id: getPulumiResourceID(this),
            props: {
                resourceProviderNamespaces: providers
                    .filter((p) => currentProps.resourceProviderNamespaces.some((ns) => ns.toLowerCase() === p.namespace?.toLowerCase()))
                    .map(({ namespace }) => namespace ?? ""),
            },
        };
    }
    async update(id, currentOutputs, newInputs) {
        const providersToAdd = newInputs.resourceProviderNamespaces.filter((nsIn) => !currentOutputs.resourceProviderNamespaces.some((nsOut) => nsIn.toLowerCase() === nsOut.toLowerCase()));
        const providersToRemove = currentOutputs.resourceProviderNamespaces.filter((nsOut) => !newInputs.resourceProviderNamespaces.some((nsIn) => nsIn.toLowerCase() === nsOut.toLowerCase()));
        const client = getClient(this);
        await Promise.all([
            ...providersToAdd.map((p) => client.providers.register(p)),
            ...providersToRemove.map((p) => client.providers.unregister(p)),
        ]);
        return {};
    }
    async delete(id, props) {
        const client = getClient(this);
        await Promise.all(props.resourceProviderNamespaces.map((ns) => client.providers.unregister(ns)));
    }
}
const getPulumiResourceID = ({ tenantId, subscriptionId, }) => `${tenantId}/${subscriptionId}`;
const getClient = ({ subscriptionId, ...provider }) => new resources.ResourceManagementClient(getAzureCredential(provider), subscriptionId);
const getAzureCredential = ({ auth, tenantId, keyAndCertPath, }) => {
    switch (auth.type) {
        case "msi":
            return new id.ManagedIdentityCredential(auth.clientId);
        case "sp": {
            return new id.ClientCertificateCredential(tenantId, auth.clientId, keyAndCertPath);
        }
        default:
            throw new Error(`Unsupported auth type "${auth.type}".`);
    }
};
//# sourceMappingURL=provider-registration.js.map